#include "string.h"
#include "sis.h"
#include "th_int.h"


static int minterm_compare(int *s1, int *s2);
static int ninput = 3;


/****************************************************** 
 * given an array of minterm indexes, sort this array *
 * in increasing order wrt number index               *
 ******************************************************/

void majsyn_sort_minterms(array_t *minterms)
{
  /* sort the array */
  array_sort(minterms, minterm_compare);
}

void majsyn_uniq_minterms(array_t *minterms)
{ /* make the array unique */
  array_uniq(minterms, minterm_compare, (void (*)()) 0); 
}


/*********************************************************************** 
 * This function itself is used as a parameter in function: array_sort *
 * It compares two minterms by the minterm index                       *
 * returns -1 if n1 is greater, 0 otherwise                            *
 ***********************************************************************/
int minterm_compare(int *s1, int *s2)
{
  /* return appropriate status */
  if (*s1 < *s2)
    return -1;
  else if (*s1 == *s2)
    return 0;
  else
    return 1;
}

/********************************************************************/
/* given the unique sorted minterm indexes of an admissible pattern */
/* return the key for this admissible pattern in hash table         */
/********************************************************************/
char *majsyn_create_key(array_t *minterms)
{
  int i, min;
  int  length;
  char *key;
  char *s;
  
  key = ALLOC(char, 50);
  s = ALLOC(char, 50);

  /* get the first minterm index */
  length = sprintf(s, "%d", array_fetch(int, minterms, 0));
  strcpy(key, s);
  for (i = 1; i < array_n(minterms); i++) {
    min = array_fetch(int, minterms, i);
    length = sprintf(s, "%d", min);
    key = strcat(key, s);
  }
  free(s);
  return key;
}

/*******************************************************/
/* return the indexes of cubes the pset_family covers, */
/* the pset_family needn't to be the unique expression */
/*******************************************************/

array_t *pfamily_minterms(pset_family f)
{
  int i, j, sum_index;
  int *sum, *bit;
  array_t *my_pfamily_minterms, *pset_minterms;
  register pset last, p;

  my_pfamily_minterms = array_alloc(int,0);


  /* initialize the sum array, which corresponds to the minterm index, m0 - m7 */
  sum = ALLOC(int, 8);
  /* initialize the bit array, which corresponds to the bit: b0 b1 b2 b3 b4 b5 */
  bit = ALLOC(int, ninput * 2);

  foreach_set(f, last, p) { 
   for (i = 0; i < ninput * 2; i++) {
      bit[i] = is_in_set(p,i)? 1 : 0; 
   }
   sum_index = 0;
    /* check variable x1 */
    if (bit[0] == 1 && bit[1] == 0)
      sum[0] = 0;  
    if (bit[0] == 0 && bit[1] == 1)
      sum[0] = 4;
    if (bit[0] == 1 && bit[1] == 1) {
      sum[0] = 0;
      sum[1] = 4;
      sum_index++;
    }
    /* check variable x2 */
    if (bit[2] == 1 && bit[3] == 0) {
      
      for (j = 0; j <= sum_index; j++)
	sum[j] += 0;  
    }
    if (bit[2] == 0 && bit[3] == 1) {
      for (j = 0; j <= sum_index; j++)
	sum[j] += 2;  
    }
    if (bit[2] == 1 && bit[3] == 1) {
      for (j = 0; j <= sum_index; j++)
	sum[j] += 0;  
      for (j = 0; j <= sum_index; j++)
	sum[sum_index + j + 1] = sum[j] + 2;  
      sum_index = (sum_index + 1) * 2 - 1;
    }
    /* check variable x3 */
    if (bit[4] == 1 && bit[5] == 0) {
      for (j = 0; j <= sum_index; j++)
	sum[j] += 0;  
    }
    if (bit[4] == 0 && bit[5] == 1) {
      for (j = 0; j <= sum_index; j++)
	sum[j] += 1;  
    }
    if (bit[4] == 1 && bit[5] == 1) {
      for (j = 0; j <= sum_index; j++)
	sum[j] += 0;  
      for (j = 0; j <= sum_index; j++)
	sum[sum_index + j + 1] = sum[j] + 1;  
      sum_index = (sum_index + 1) * 2 - 1;
    }
  /* find out the minterms */
  pset_minterms = array_alloc(int,0);
  for (i = 0; i <= sum_index; i++)
    array_insert_last(int, pset_minterms, sum[i]);
  array_append(my_pfamily_minterms, pset_minterms);
  array_free(pset_minterms);
  }

  /* sort the minterm array in descending order */
  majsyn_sort_minterms(my_pfamily_minterms);
  /* make the minterm array unique */
  majsyn_uniq_minterms(my_pfamily_minterms);

  return my_pfamily_minterms;
}


/****************************************************************/
/* given a pset_family, which consists of '-' (11) in the pset, */ 
/* return a pset_family which is functionally equivalent to the */
/* originalthe one, but pset_family only consists of minterms,  */
/* no '-' (11)                                                  */
/****************************************************************/

pset_family pfamily_all_minterms(pset_family f)
{
  int i, m_index;
  pset p;
  pset_family new_f;
  array_t *minterms;

  new_f = sf_new(8, 2 * ninput);
  /* get the minterm indexes that f covers */
  minterms = pfamily_minterms(f);
  for (i = 0; i < array_n(minterms); i++) {
    m_index = array_fetch(int, minterms, i);
    p = minterm_by_index(m_index);
    new_f = sf_addset(new_f, p);
    set_free(p);
  }

  return new_f;
}

/********************************************************/
/* given a minterm index, return the corresponding pset */
/********************************************************/
pset minterm_by_index(int n)
{ 
  pset p;

  if (n < 0 || n > 8)
    fail("minter_by_index: Minterm index not in range");

  p = set_new(ninput * 2);
  switch (n) {
  case 0:
    set_insert(p, 2 * 0);
    set_insert(p, 2 * 1);
    set_insert(p, 2 * 2);
    break;
  case 1:
    set_insert(p, 2 * 0);
    set_insert(p, 2 * 1);
    set_insert(p, 2 * 2 + 1);
    break;
  case 2:
    set_insert(p, 2 * 0);
    set_insert(p, 2 * 1 + 1);
    set_insert(p, 2 * 2);
    break;
  case 3:
    set_insert(p, 2 * 0);
    set_insert(p, 2 * 1 + 1);
    set_insert(p, 2 * 2 + 1);
    break;
  case 4:
    set_insert(p, 2 * 0 + 1);
    set_insert(p, 2 * 1);
    set_insert(p, 2 * 2);
    break;
  case 5:
    set_insert(p, 2 * 0 + 1);
    set_insert(p, 2 * 1);
    set_insert(p, 2 * 2 + 1);
    break;
  case 6:
    set_insert(p, 2 * 0 + 1);
    set_insert(p, 2 * 1 + 1);
    set_insert(p, 2 * 2);
    break;
  case 7:
    set_insert(p, 2 * 0 + 1);
    set_insert(p, 2 * 1 + 1);
    set_insert(p, 2 * 2 + 1);
    break;
  default:
    fail("minter_by_index: Minterm index not in range");
  }

  return p;
}


/*******************************************************/
/* find a given element(same value, integer) in array  */
/* return: 1 find this element                         */
/*         0 didn't find this element                  */
/* note: make sure the elements in array are integer   */
/*******************************************************/

int majsyn_array_find_element(array_t *array, int n)
{ 
  int i;
  
  for (i = 0; i < array_n(array); i++) {
    if (n == array_fetch(int, array, i))
      return 1;
  }
  return 0;
}

/*****************************************************/
/* Compare if two array_t (both integer type) have   */
/* exactly the same elements.                        */
/* return: 1 same                                    */
/*         0 not same                                */
/*****************************************************/

int two_int_array_same(array_t *a1, array_t *b1)
{
  int i, a_value;
  
  if (array_n(a1) != array_n(b1))
    return 0; // not same
  for (i = 0; i < array_n(a1); i++) {
    a_value = array_fetch(int, a1, i);
    if (majsyn_array_find_element(b1, a_value) == 0)
      return 0; // not same
  }
  return 1; // same
}


/********************************************************************/
/* Given an array, exclude_patterns, whose elements are also        */
/* array of integer values. Check if another array of integer       */
/* values,this_pattern, is an element of exlude_patterns            */
/* return: 1: this pattern IS an element of exlude_patterns         */
/*         0: ............ NOT .........                            */
/********************************************************************/

int pattern_is_element(array_t *exclude_patterns, array_t *this_pattern)
{
  int i;
  array_t *exclude;

  if (array_n(exclude_patterns) == 0) 
    return 0; /* NOT an element */
  for (i = 0; i < array_n(exclude_patterns); i++) {
    /* get out one exclude pattern */
    exclude = array_fetch(array_t *, exclude_patterns, i); 
    if (two_int_array_same(exclude, this_pattern)) /* this two patterns are the same */
      return 1; /* IS an element */
  }
  return 0; /* NOT an element */
}

/****************************************************************************/
/* Given an array of pset_family which consists of all excluded pset_familys,*/
/* create an array for it, whose elements are array of integer indexes for  */
/* each of the pset_family                                                  */
/****************************************************************************/
array_t *transform_exclude_pfamilys(array_t *f_array)
{
  array_t *exclude_patterns;
  array_t *minterms;
  int i;

  exclude_patterns = array_alloc(array_t *, 0);
  
  for (i = 0; i < array_n(f_array); i++){ /* get the minterm indexes for each pset_family */
    minterms = pfamily_minterms(array_fetch(pset_family, f_array, i));
    array_insert_last(array_t *, exclude_patterns, minterms);

  }
  return exclude_patterns;
}



/*************************************************/
/* given the on_sets, compute the off_sets       */
/* return the off_sets saved in pset_family      */
/*************************************************/
pset_family comput_off_from_on(pset_family f_on)
{
  int i;
  pset p;
  pset_family f_off;
  array_t *minterms_on;

  minterms_on = pfamily_minterms(f_on); 
  if (array_n(minterms_on) >= 8)
    f_off = NULL;
  else {
    f_off = sf_new(8, 2 * ninput);
    for( i = 0; i < 8; i++)
      if (majsyn_array_find_element(minterms_on, i) == 0) {
	/* this minterm not in on_set */
	p = minterm_by_index(i);
	f_off = sf_addset(f_off, p);
	set_free(p);
      }
  }
  return f_off;
}      

/**********************************************************/
/* this function computes the must_set based on given     */
/* three pset_familys: f0, f1, f2.  Find must0_set, if    */
/* phase = 0, find must1_set if phase = 1, return the     */
/* must_set in a pset_family.                             */
/* where f0 is the original node on_set, f1 is the        */
/* on-sets fset_family got first from our synthesis       */
/* method, f2 is the on-set got second from our synthesis.*/
/* Note: the order of f0, f1, f2 makes difference         */
/**********************************************************/
pset_family find_must_set(pset_family f0, 
			  pset_family f1,
			  pset_family f2,
			  int phase)
{
  int i, m, count;
  pset p;
  pset_family must_set;
  array_t *f0_minterms, *f1_minterms, *f2_minterms;

  if ((phase != 0) && (phase !=1))
    fail("find_must_set: phase is not 0 or 1");
  if (f0 == NULL) {
    must_set = NULL;
    return must_set;
  }
  if (f1 == NULL)
    fail("find_must_set: f1 is empty");
  must_set = sf_new(8, 2 * ninput);
  /* get the minterm indexes of f0, f1 and f2 in correct phase */
  if (phase == 1) {
    f0_minterms = pfamily_minterms(f0);
    f1_minterms = pfamily_minterms(f1);
    if (f2 != NULL)
      f2_minterms = pfamily_minterms(f2);
  }
  else if (phase == 0) {
    f0_minterms = pfamily_minterms(comput_off_from_on(f0));
    f1_minterms = pfamily_minterms(comput_off_from_on(f1));
    if (f2 != NULL)
      f2_minterms = pfamily_minterms(comput_off_from_on(f2));
  }    
  for(i = 0; i < array_n(f0_minterms); i++) {
    count = 0; 
    m = array_fetch(int, f0_minterms, i); /* get the minterm index */
    if (majsyn_array_find_element(f1_minterms, m)) /* this minterm exist in f1_minterms */
      count++;
    if (f2 != NULL)
       if (majsyn_array_find_element(f2_minterms, m)) /* this minterm exist in f2_minterms */
	 count++;
    if ((f2 == NULL) && (count == 0)) {/* this minterm must be in must_set */
      p = minterm_by_index(m);
      must_set = sf_addset(must_set, p);
      set_free(p);
    }
    else if (f2 != NULL) {
      if (count < 1)
	fail("find_must_set: can't find must_set");
      if (count == 1) {/* this minterm must be in must_set */
	p = minterm_by_index(m);
	must_set = sf_addset(must_set, p);
	set_free(p);
      }
    }
  }

  return must_set;
}

/*************************************************/
/* given an array of minterm indexes, return the */
/* corresponding pset_family (not simplified)    */
/*************************************************/
pset_family minterm_indexes_to_pfamily(array_t *result)
{
  int i, m;
  pset_family f;
  pset p;

  if (array_n(result) <=0)
    f = NULL;
  else {  
    f = sf_new(8, 2 * ninput);
    for (i = 0; i < array_n(result); i++){
      m = array_fetch(int, result, i);
      p = minterm_by_index(m);
      f = sf_addset(f, p);
      set_free(p);
    }
  }
  return f;
}

/*******************************************************/
/* given the original network, original node n,        */
/* and three pset_familys: f1, f2 and f3               */
/* (all of them are non-empty, and simplified), which  */
/* are got from our majority synthesis method, create  */
/* new nodes: n1, n2, n3 which has on-set pset_family  */
/* f1, f2, and f3 respectively. And also create new    */
/* node n0, which is a majority function:              */
/* n0 = f1 f2 + f1 f3 + f2 f3. Add nodes n1, n2 and n3 */
/* into the original network, and replace original n   */
/* with n0 in original network                         */
/*******************************************************/

void majsyn_replace_majnode (network_t *network, node_t *n, 
			     pset_family *f1, pset_family *f2,
			     pset_family *f3) 
{ 
  node_t *n1, *n2, *n3;
  node_t *tmp_node;
  node_t *clit, *g1, *g2, *g3, *t;

  /* create a node n1 for f1 */
  n1 = node_dup(n);
  n1->F = sf_copy(n1->F, *f1);
  network_add_node(network, n1);

  /* create a node n2 for f2 */
  n2 = node_dup(n);
  n2->F = sf_copy(n2->F, *f2);
  network_add_node(network, n2);

  /* create a node n3 for f3 */
  n3 = node_dup(n);
  n3->F = sf_copy(n3->F, *f3);
  network_add_node(network, n3);

  /* create a node n0 for f_original */ 

  g1 = node_constant(1);
  clit = node_literal(n1, 1);
  t = node_and(g1, clit);
  node_free(clit);
  node_free(g1);
  g1 = t; /* g1 = n1 */

  clit = node_literal(n2, 1);
  t = node_and(g1, clit);
  node_free(clit);
  node_free(g1);
  g1 = t; /* g1 = n1 n2 */

  g2 = node_constant(1);
  clit = node_literal(n1, 1);
  t = node_and(g2, clit);
  node_free(clit);
  node_free(g2);
  g2 = t; /* g2 = n1 */

  clit = node_literal(n3, 1);
  t = node_and(g2, clit);
  node_free(clit);
  node_free(g2);
  g2 = t; /* g2 = n1 n3 */

  g3 = node_constant(1);
  clit = node_literal(n2, 1);
  t = node_and(g3, clit);
  node_free(clit);
  node_free(g3);
  g3 = t; /* g3 = n2 */

  clit = node_literal(n3, 1);
  t = node_and(g3, clit);
  node_free(clit);
  node_free(g3);
  g3 = t; /* g2 = n2 n3 */

  tmp_node = node_or(g1, g2);
  tmp_node = node_or(tmp_node, g3);

  node_replace(n, tmp_node); 

}


/************************************************************************/
/* create the hash tables which contains all admissible patterns        */
/* table_1: for all admissible patterns of 'xi' format(total:6 patterns)*/
/* table_2: for all admissible patterns of 'xixj' or 'xi+xj' format     */
/* (total: 24 patterns)                                                 */
/* table_3: for all admissible patterns of 'T' shape(total: 8 patterns) */
/************************************************************************/

st_table *maj_table_1(void) 
{
  st_table *table_1;
  pset p;
  pset_family f;
  char *pbv1();
  char *name;

  /* table_1: for all admissible patterns of 'xi' format(total:6 patterns)*/
  table_1 = st_init_table(strcmp, st_strhash); 
  /* f = x1 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("4567");
  st_insert(table_1, name, (char *) f);

  /* f = x1' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("0123");
  st_insert(table_1, name, (char *) f);

  /* f = x2 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("2367");
  st_insert(table_1, name, (char *) f);

  /* f = x2' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("0145");
  st_insert(table_1, name, (char *) f);

  /* f = x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("1357");
  st_insert(table_1, name, (char *) f);

  /* f = x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("0246");
  st_insert(table_1, name, (char *) f);

 return table_1;
}

st_table *maj_table_2(void) 
{
  st_table *table_2;
  pset p;
  pset_family f;
  char *pbv1();
  char *name;

  /* table_1: for all admissible patterns of */
  /* 'xi xj' or 'xi + xj' format(total:24 patterns)*/
  table_2 = st_init_table(strcmp, st_strhash); 
  /* f = x1 + x2 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("234567");
  st_insert(table_2, name, (char *) f);

  /* f = x1' + x2 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("012367");
  st_insert(table_2, name, (char *) f);

  /* f = x1 + x2' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("014567");
  st_insert(table_2, name, (char *) f);

  /* f = x1' + x2' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("012345");
  st_insert(table_2, name, (char *) f);

  /* f = x1 + x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("134567");
  st_insert(table_2, name, (char *) f);

  /* f = x1' + x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("012357");
  st_insert(table_2, name, (char *) f);

  /* f = x1 + x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("024567");
  st_insert(table_2, name, (char *) f);

  /* f = x1' + x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("012346");
  st_insert(table_2, name, (char *) f);

  /* f = x2 + x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 );
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("123567");
  st_insert(table_2, name, (char *) f);

  /* f = x2' + x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2 );
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("013457");
  st_insert(table_2, name, (char *) f);

  /* f = x2 + x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 );
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("023467");
  st_insert(table_2, name, (char *) f);

  /* f = x2' + x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2 );
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("012456");
  st_insert(table_2, name, (char *) f);

  /* f = x1 x2 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("67");
  st_insert(table_2, name, (char *) f);

  /* f = x1' x2 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("23");
  st_insert(table_2, name, (char *) f);

  /* f = x1 x2' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("45");
  st_insert(table_2, name, (char *) f);

  /* f = x1' x2' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("01");
  st_insert(table_2, name, (char *) f);

  /* f = x1 x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("57");
  st_insert(table_2, name, (char *) f);

  /* f = x1' x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("13");
  st_insert(table_2, name, (char *) f);

  /* f = x1 x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("46");
  st_insert(table_2, name, (char *) f);

  /* f = x1' x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("02");
  st_insert(table_2, name, (char *) f);

  /* f = x2 x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("37");
  st_insert(table_2, name, (char *) f);

  /* f = x2' x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("15");
  st_insert(table_2, name, (char *) f);

  /* f = x2 x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("26");
  st_insert(table_2, name, (char *) f);

  /* f = x2' x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("04");
  st_insert(table_2, name, (char *) f);

  return table_2;
}



st_table *maj_table_3(void) 
{
  st_table *table_3;
  pset p;
  pset_family f;
  char *pbv1();
  char *name;

  /* table_3: for all admissible patterns of 'T' shape(total:8 patterns)*/
  table_3 = st_init_table(strcmp, st_strhash); 
  /* f = x1 x2 + x1 x3 + x2 x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("3567");
  st_insert(table_3, name, (char *) f);

  /* f = x1' x2 + x1' x3 + x2 x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("1237");
  st_insert(table_3, name, (char *) f);

  /* f = x1 x2' + x1 x3 + x2' x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("1457");
  st_insert(table_3, name, (char *) f);

  /* f = x1 x2 + x1 x3' + x2 x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("2467");
  st_insert(table_3, name, (char *) f);

  /* f = x1' x2' + x1' x3 + x2' x3 */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("0135");
  st_insert(table_3, name, (char *) f);

  /* f = x1' x2 + x1' x3' + x2 x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("0236");
  st_insert(table_3, name, (char *) f);

  /* f = x1 x2' + x1 x3' + x2' x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0 +1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("0456");
  st_insert(table_3, name, (char *) f);

  /* f = x1' x2' + x1' x3' + x2' x3' */
  f = sf_new(8, 2 * ninput);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  set_insert(p, 2 * 2 + 1);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 1 + 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  p = set_new(ninput * 2);
  set_insert(p, 2 * 0);
  set_insert(p, 2 * 0 + 1);
  set_insert(p, 2 * 1);
  set_insert(p, 2 * 2);
  f = sf_addset(f, p);
  set_free(p);
  /* save this pset_family in table */
  name = util_strsav("0124");
  st_insert(table_3, name, (char *) f);

  return table_3;
}


/*********************************************************/
/* check if a pset_family matches any admissible pattern */
/* return: 0: Not an admissible pattern                  */
/*         1: Is an admissible pattern in t1             */
/*         2: Is an admissible pattern in t2             */
/*         3: Is an admissible pattern in t3             */
/*********************************************************/

int check_is_majority(pset_family f, st_table *t1, st_table *t2, st_table *t3)
{
  int total, is_maj = 0;
  array_t *minterms;
  char *st_key;

  minterms = pfamily_minterms(f);
  /* get the total number of minterms covered by the pset_family */
  total = array_n(minterms);
  if (total == 2) {
    /* create key here */
    st_key = majsyn_create_key(minterms);
    is_maj = st_lookup(t2, st_key, NIL(char *));
    FREE(st_key);
    if (is_maj == 1) 
      return 2;
    else
      return 0; 
  } 
  else if (total == 4) {
    /* create key here */
    st_key = majsyn_create_key(minterms);
    /* check existance */
    is_maj = st_lookup(t1, st_key, NIL(char *));
    if (is_maj == 1) {
      FREE(st_key);
      return 1;
    }
    else {
      is_maj = st_lookup(t3, st_key, NIL(char *));
      FREE(st_key);
      if (is_maj == 1) 
	return 3;
      else
	return 0;
    }
  }
  else if (total == 6) {
    /* create key here */
    st_key = majsyn_create_key(minterms);
    /* check existance */
    is_maj = st_lookup(t2, st_key, NIL(char *));
    FREE(st_key);
    if (is_maj == 1)
      return 2;
    else
      return 0; 
  }
  else 
    return 0;
}



majsyn_options_t *majsyn_options_init(int argc, char **argv)
{
  int c;
  majsyn_options_t *opt;
  
  opt = ALLOC(majsyn_options_t, 1);

  /* default values */
  opt->m = -1; /* collapse fanin restriction (default no collapsing) */
  opt->help = FALSE;

  /* parse options */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "n:h")) != EOF) {
    switch (c) {
    case 'n': opt->m = atoi(util_optarg); break;
    case 'h': opt->help = TRUE; break;
    default: opt->help = TRUE; break;
    }
  }
 
  return opt;
}


/*******************************
 * free options data structure *
 *******************************/
void majsyn_options_free(majsyn_options_t *opt)
{
  if (opt == NIL(majsyn_options_t))
    return;

  FREE(opt);
}
