#include <math.h>
#include "sis.h"
#include "th_int.h"

static int maptpl_usage ARGS((void));
static st_table *my_level_table; /* For the levelize routines */
static int level_cmp_decrease ARGS((char **, char **));
static int my_assign_level ARGS((node_t *)); 
static void new_fanout_nodes ARGS((network_t *, node_t *, int));

int com_maptpl(network_t **network, int argc, char **argv)
{
  int c, i, max_fo, violate = 0;
  array_t *node_vec;
  node_t *node, *p;
  lsGen gen;
  st_table *table_1, *table_2, *table_3;  

  ///////////////for collecting experimental results only////////////////////////////
    printf("network: %s\n", network_name(*network));
    printf("before: %d\n", network_num_internal(*network));
  ///////////////////////////////////////////////////////////////////////////////////

  /* set the default value for the maximum fanout restriction */
  max_fo = 3;

  util_getopt_reset();

   while ((c = util_getopt(argc, argv, "o:h")) != EOF) {
    switch(c) {
    case 'o': max_fo = atoi(util_optarg); break;
    case 'h':
      return maptpl_usage();
    default:
      return maptpl_usage();
    }
  }

  my_level_table = st_init_table(st_ptrcmp, st_ptrhash);
  foreach_node(*network, gen, p) {
    (void) st_insert(my_level_table, (char *)p, (char *)0);
  }
  
  foreach_primary_output(*network, gen, p) {
    (void) my_assign_level(p); //recursively assign the levels
  }

  node_vec = network_dfs_from_input(*network);
  /* sort the node in network from highest level to lowest level */
  /* the sorted nodes are saved in node_vec */
  array_sort(node_vec, level_cmp_decrease);

  for (i = 0; i < array_n(node_vec); i++) {
    node = array_fetch(node_t *,node_vec,i);
    if ((node_type(node) != PRIMARY_OUTPUT) && (node_type(node) != PRIMARY_INPUT)) 
      { 
/* 	//debug */
/* 	printf("%s, fanout = %d\n",node_name(node), node_num_fanout(node)); */
/* 	node_t *fanout; */
/* 	foreach_fanout(node, gen, fanout)  */
/* 	  printf("%s", node_name(fanout)); */
/* 	printf("\n"); */
/* 	//debug */
	if (node_num_fanout(node) > max_fo) {
	  violate++;
	  new_fanout_nodes(*network, node, max_fo);
	}
      }
  }
  //  if (violate > 0)
  //    printf("Network %s: %d nodes\n", network_name(*network), violate);
  array_free(node_vec);
  st_free_table(my_level_table);
  
 ///////////////for collecting experimental results only////////////////////////////
    printf("after: %d\n", network_num_internal(*network));
 ///////////////////////////////////////////////////////////////////////////////////


  return 0;
}

static void new_fanout_nodes(network_t *network, node_t *node, int max_fo)
{
  node_t *dup_node, *fanout, *new_node;
  int i, j, fo_num, upper, fo_index, fi_index;
  lsStatus status;
  lsGeneric *data;
  lsHandle *itemHandle;
  array_t *node_vec;
  array_t *fo_array, *reconnect_nodes, *fiindex_array; 

  /* save the fanouts of the node in a node array */
  fo_num = node_num_fanout(node);
  node_vec = array_alloc(node_t *, 0);
  for (i = 0; i < fo_num; i++)
    array_insert_last(node_t *, node_vec, node_get_fanout(node,i));
  //  printf("the node to be duplicated:\n");
  //  node_print(sisout, node);

  /* count the total number copies of nodes needed for this node */
  upper = ceil(fo_num/(double)max_fo);
  if (upper < 2)
    return;

  /* initialize the arrays */
  fo_array = array_alloc(node_t *, 0);
  reconnect_nodes = array_alloc(node_t *, 0);
  fiindex_array = array_alloc(int, 0);
  for (i = 1; i < upper; i++) {
    /* duplicate the node that has more than max_fo fanout nodes */
    dup_node = node_dup(node); 
    /* add this new node into the network */
    node_assign_name(dup_node);
    node_assign_short_name(dup_node);
 
    network_add_node(network, dup_node); 
    for (j = 0; j < max_fo; j++) { 
	  fo_index = max_fo * i + j;
	  if (fo_index >= fo_num)
	    break;
	  array_insert_last(node_t *, fo_array, dup_node);
	  array_insert_last(node_t *, reconnect_nodes, array_fetch(node_t *, node_vec, fo_index));
	  fi_index = node_get_fanin_index(node_get_fanout(node,fo_index), node);
	  array_insert_last(int, fiindex_array, fi_index);
    }
  }
  
  /* remove the original fanout lists */
  for (i = 0; i < array_n(node_vec); i++) {
    fanout = array_fetch(node_t *, node_vec, i);
    fanin_remove_fanout(fanout);
  }
  
  /* substitute some fanins of the node by the newly duplicated */
  /* node, therefore, reduce the number of fanout nodes of a node */
  for(i = 0; i < array_n(fiindex_array); i++) {
    new_node = array_fetch(node_t *, reconnect_nodes, i);
    fi_index = array_fetch(int, fiindex_array, i);
    new_node->fanin[fi_index] = array_fetch(node_t *, fo_array, i);
  }

  /* update the fanout list of the original node */
  for (i = 0; i < array_n(node_vec); i++) {
    fanout = array_fetch(node_t *, node_vec, i);
    fanin_add_fanout(fanout);
  }
  array_free(node_vec);
  array_free(fiindex_array);
  array_free(fo_array);
  array_free(reconnect_nodes);
}

static int my_assign_level(node_t *node) 
{
  node_t *fanin;
  int i, level, new_level;
  char *dummy;
  
  (void) st_lookup(my_level_table, (char *)node, &dummy);
  new_level = (int)dummy;
  if (node->type == PRIMARY_INPUT || new_level > 0) {
    return new_level ;
  } else {
    level = 0;
    foreach_fanin(node, i, fanin) {
      level = MAX(level, my_assign_level(fanin));
    }
  }
  level++;
  (void) st_insert(my_level_table, (char *)node, (char *)level);

  return level;
}

static int level_cmp_decrease(char **p1, char **p2) 
{
  int level1, level2;
  char *dummy;
  
  (void) st_lookup(my_level_table, *p1, &dummy);
  level1 = (int) dummy;
  (void) st_lookup(my_level_table, *p2, &dummy);
  level2 = (int) dummy;
  return (level2 - level1); //rui
}


/*****************************************/
/* usage information of majcheck command */
/*****************************************/
static int maptpl_usage(void)
{
  fprintf(siserr, "usage: majsyn [-o int1]\n");
  fprintf(siserr, "    map the majority/minority network to tpl nanotechnology.\n");
  fprintf(siserr, "    -o int1: the maximum allowed fanout.\n");
  return 0;
}
