#include <stdio.h> 
#include <math.h>
#include "sis.h"
#include "../stg/stg_int.h"

static int my_states_usage ARGS(());
static int my_log2 ARGS((int));

int com_my_states(network_t **network, int argc, char **argv) 
{
  int c;
  graph_t *stg;
  int ns, num_states, code_length, help = 0;
  char *code, *filename;
  lsGen gen;
  vertex_t *s;
  network_t *new;
  FILE *p_file;
  
  num_states = 0;
  filename = ALLOC(char, 100);
  code_length = 0;
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "n:f:c:h")) != EOF) {
    switch (c) {
    case 'n': 
      num_states = atoi(util_optarg); 
      filename = util_strsav("/home/sis/genesis/experiments/tmp_states.txt");
      break;
    case 'f': 
      filename = util_strsav(util_optarg); //TODO: append '.experiments' as prefix
      break; 
    case 'c':
      code_length = atoi(util_optarg);
      break;
    case 'h': help = 1; break;
    default: help = 1; break;
    }
  }
  
  if (help  == 1) return my_states_usage();
  if (num_states == 0) return my_states_usage();
  if (code_length <= 0)
    code_length = my_log2(num_states);
  //DEBUG
  //  printf("code_length = %d\n",code_length);
  //end DEBUG

    error_init();
    if (network_stg(*network) == NIL(graph_t)) {
      if (filename != NIL(char))
	FREE(filename);
      (void) fprintf(siserr, "There is no stg\n");
      return 1;
    }
    //DEBUG
    //    printf("entering com_my_states, num_states = '%d', filename = '%s'.\n", num_states, filename);
    //end DEBUG

    stg = network_stg(*network);
    if ((p_file = fopen(filename, "r")) == NULL) 
      fail("com_my_states: unable to open file handle for reading."); 

    ns = stg_get_num_states(stg); 
    if (ns != num_states)
      (void) fprintf(siserr, "my_set_states: number of states is inconsistent.\n");
    
    /* code_length is the number of bits needed to encode a state */
    code = ALLOC(char, code_length + 1);
    stg_foreach_state(stg, gen, s) {
      fgets(code, code_length + 1, p_file); 
      fseek(p_file, 1L,1);
      stg_set_state_encoding(s, code);
    }
    fclose(p_file);
    FREE(code); 
    
    //DEBUG 
    //stg_foreach_state(stg, gen, s) {
    //printf("states: %s\n", stg_get_state_encoding(s));
    //}
    //end DEBUG
    
    new = stg_to_network(stg, 0);
    network_set_name(new, network_name(*network));
    stg_set_network_pipo_names(new, network_stg(*network));
    
    /* Add the clocks and set the latch type and control */
    //    stg_network_add_clock(new, stg);
    //   stg_network_connect_clock(new, stg);

    network_set_stg(*network, NIL(graph_t));
    network_free(*network);
    *network = new;

    if (filename != NIL(char))
    FREE(filename);
    return 0;
}



static int my_log2(int n) {
  int i;
  double j;

  if (n <= 0) {
    printf("Help! values is %d, must be positive!\n", n);
    abort();
  }
  
  j = log((double) n)/ log(2);
  i = (int) ceil(j);

  return i;
}

/*******************
 * my_states usage *
 *******************/
static int my_states_usage(void)
{
  (void) fprintf(siserr, "usage: my_states -n num1 [-c num2][-f filename]\n");
  fprintf(siserr, "The encoding of the states are given in a file in binary format. The states of the stg network are set according to the values written in the file.\n");
  fprintf(siserr, "   -n num1:\tGives the total number of states of the FSM. This option is a requirment\n");
  fprintf(siserr, "   -c num2:\tGives the code length of the states. (default: ceil(log2(num1)))\n");
  fprintf(siserr, "   -f file:   \tRead the state assignment from the file.\n");
  fprintf(siserr, "              \tBy default, read from file 'tmp_states.txt'.\n");
  fprintf(siserr, "   -h:        \tFor help information. \n");  
  return 1;
}

