/*=========================================================================*
 *	File    : thatpg_cubes.c
 *
 *	Author  : Pallav Gupta
 *	Created : Tue Dec 30 22:28:16 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thatpg_cubes.c,v 1.2 2004/01/19 07:17:32 pgupta Exp $ */
#include "sis.h"
#include "th_int.h"

/* calculate the singular covers for each gate in the network */
int thatpg_calc_scovers(network_t *network)
{
  node_t *n;
  lsGen gen;
  st_table *scovers;
  char *key;
  thnode_t *th;
  st_generator *sgen;

  if (is_threshold_network(network) != TH_SUCCESS)
    return TH_FAIL;

  scovers = st_init_table(strcmp, st_strhash);
  foreach_node(network, gen, n) {
    if (node_type(n) != INTERNAL) continue;
    th = (thnode_t *) n->th;
    th->scovers = thnode_scovers(n, scovers);
  }
  st_foreach_item(scovers, sgen, (char **) &key, (char **) 0)
    st_delete(scovers, &key, (char **) 0);
  st_free_table(scovers);
  
  return TH_SUCCESS;
}

/* calculate the propagation cubes for each gate in the network */
int thatpg_calc_pcubes(network_t *network)
{
  node_t *n;
  lsGen gen;
  st_table *pcubes;
  char *key;
  thnode_t *th;
  st_generator *sgen;

  if (is_threshold_network(network) != TH_SUCCESS)
    return TH_FAIL;

  pcubes = st_init_table(strcmp, st_strhash);
  foreach_node(network, gen, n) {
    if (node_type(n) != INTERNAL) continue;
    th = ((thnode_t *) n->th);
    th->pcubes = thnode_pcubes(n, pcubes);
  }

  st_foreach_item(pcubes, sgen, (char **) &key, (char **) 0)
    st_delete(pcubes, &key, (char **) 0);
  st_free_table(pcubes);

  return TH_SUCCESS;
}

