/*=========================================================================*
 *	File    :	thcom.c
 *
 *	Author  :	Pallav Gupta
 *	Created :	Sun Sep 14 20:44:19 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/


/* $Id: thcom.c,v 1.19 2006/03/30 22:28:17 pgupta Exp $ */


#include "sis.h"
#include "th_int.h"

/* declare all usage functions as static (local scope) */

static int thconv_usage ARGS(());
static int thprint_usage ARGS(());
static int thprint_stats_usage ARGS(());


/* Add all the threshold commands to the command table */
void init_threshold(void)
{
  /* create a new entry for each command */
  /* synthesis commands */
  com_add_command("thsyn", com_thsyn, 1 /* changes network */);
  com_add_command("thsyn_red_remove", com_thsyn_with_red_removal, 1 /* changes
									network */);
  com_add_command("thconv", com_thconv, 1);

  /* majority synthesis commands */
  com_add_command("majsyn", com_majsyn, 1 /*changes network */);
  com_add_command("majsyn_red_remove", com_majsyn_with_red_removal, 1 /* changes
									 network */);
  com_add_command("majcheck", com_majcheck, 0 /*no changes in network */);

 /* technology mapping commands */
  com_add_command("thbuffer", com_thbuffer, 1 /*changes network */);
  com_add_command("maptpl", com_maptpl, 1 /*changes in network */);

  /* state encoding commands */
  com_add_command("my_states", com_my_states, 1 /*changes network*/);

  /* simulation commands */
  com_add_command("thsim", com_thsim, 0);

  /* testing/fault sim commands */
  com_add_command("thfsim", com_thfsim, 0);
  com_add_command("thatpg", com_thatpg, 0);

  /* misc commands */
  com_add_command("thprint", com_thprint, 0);
  com_add_command("thprint_stats", com_thprint_stats, 0);
  com_add_command("thprint_faults", com_thprint_faults, 0);
  com_add_command("thvcg", com_thvcg, 0);
  com_add_command("thdot", com_thdot, 0);
  com_add_command("invcheck", com_invcheck, 0);

  
  /* factorization based on threshold logic commands */
  com_add_command("thfactor", com_thfactor, 1 /*changes network*/);

  /* factorization based on majority logic commands */
  com_add_command("majfactor", com_majfactor, 1 /*changes network*/);
}

/* called when SIS is exiting */
void end_threshold(void)
{
   /* nothing to do */
}

/* prints all the lines (checkpoint theorem) that must be targetted for test
   generation */
int com_thprint_faults(network_t **network, int argc, char **argv) 
{
  int i, c;
  char *filename;
  FILE *fp = sisout;
  lsGen gen, gen1;
  node_t *n, *fanin, *fanout;
  char *s, *s1;

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "f:h")) != EOF) {
    switch(c) {
    case 'f':
      filename = util_strsav(util_optarg);
      if ((fp = com_open_file(filename, "w", 0, 0)) == NIL(FILE)) {
	fprintf(siserr, "Unable to open file '%s' to write\n", filename);
	free(filename);
	return TH_FAIL;
      }
      break;
    case 'h': goto exit;
    default: goto exit;
    }
  }

  foreach_node(*network, gen, n) {
      s = util_strsav(node_name(n));
      if (s[0] == '{')   
        strcpy(s, &s[1]);
      if (s[strlen(s)-1] == '}')
        s[strlen(s)-1] = 0;

    switch (node_type(n)) {

    case PRIMARY_INPUT:
      foreach_fanout(n, gen1, fanout) {
        s1 = util_strsav(node_name(fanout));
	if (s1[0] == '{')  
          strcpy(s1, &s1[1]);
        if (s1[strlen(s1)-1] == '}')
          s1[strlen(s1)-1] = 0;
	// no need to target PI nodes that directly go to the output
        if (node_function(fanout) != NODE_PO) {
	  if (node_input_phase(fanout,  n) == NEG_UNATE) {
	    fprintf(fp, "%s %s_bar %s\n", s1, s, "SA0");
            fprintf(fp, "%s %s_bar %s\n", s1, s, "SA1");
	  }
	  else {
	    fprintf(fp, "%s %s %s\n", s1, s, "SA0");
            fprintf(fp, "%s %s %s\n", s1, s, "SA1");
	  }
	}
        free(s1);
      }
      break;
    case PRIMARY_OUTPUT:
      break;
    case INTERNAL:
      foreach_fanin(n, i,  fanin) {
        s1 = util_strsav(node_name(fanin));
	 if (s1[0] == '{')
          strcpy(s1, &s1[1]);
        if (s1[strlen(s1)-1] == '}')
          s1[strlen(s1)-1] = 0;
	if ((node_type(fanin) != PRIMARY_INPUT) && (node_num_fanout(fanin) > 1)) {
	  if (node_input_phase(n, fanin) == NEG_UNATE) {
            fprintf(fp, "%s %s_bar %s\n", s, s1, "SA0");
            fprintf(fp, "%s %s_bar %s\n", s, s1, "SA1");
	  }
	  else {
	    fprintf(fp, "%s %s %s\n", s, s1, "SA0");
            fprintf(fp, "%s %s %s\n", s, s1, "SA1"); 
	  }
	}
        free(s1);
      }
      break;
    default:
      fail("encountered illegal node type for fault list generation");
    }
    free(s);
  }

  if (fp != sisout) {
    fclose(fp);
    free(filename);
  }

  return TH_SUCCESS;

exit:
  fprintf(siserr, "usage: thprint_faults [-f <file>] [-h]\n");
  fprintf(siserr, "    -f <file>\t Dump faults to file.\n");
  return TH_FAIL;
}




 /* perfrom one-to-one mapping */
int com_thconv(network_t **network, int argc, char **argv)
{
  int alim, olim, c;
  thsyn_options_t *opt;
  
  opt = ALLOC(thsyn_options_t, 1);
  /* default values */
  opt->max_fanin = 6; // rui: changed default value from 3 to 6;
  opt->delta_on = 0;
  opt->delta_off = 1;
  opt->filename = util_strsav("thconv_ampl");
  opt->help = FALSE;

  alim = olim = TH_OTO_FANIN;

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "a:o:n:f:h")) != EOF) {
    switch(c) {
    case 'a':
      alim = atoi(util_optarg); break;
    case 'o':
      olim = atoi(util_optarg); break;
    case 'n': 
      opt->delta_on = atoi(util_optarg); break;
    case 'f': 
      opt->delta_off = atoi(util_optarg); break;
    case 'h':
      return thconv_usage();
    default:
      return thconv_usage();
    }
  }

  if (alim >= olim)
    opt->max_fanin = alim;
  else 
    opt->max_fanin = olim;


  if (alim < TH_OTO_FANIN || olim < TH_OTO_FANIN)
    return thconv_usage();
  else
    th_mapoto(*network, alim, olim, opt);

  /*  this is now a valid threshod network */
  (*network)->threshold = 1;

  return TH_SUCCESS;
}

/*  thconv usage */
static int thconv_usage(void)
{
  fprintf(siserr, "usage: thconv [-a and] [-o or] [-n num1] [-f num2]\n");
  fprintf(siserr, "    -a and \tAnd gate with fanin limit 'and' (default is 2)\n");
  fprintf(siserr, "    -o or \tOr gate with fanin limit 'or' (default is 2)\n");
  fprintf(siserr, "    -n num1:\tAssign the integer num2 to delta_on, default value is 0.\n");
  fprintf(siserr, "    -f num2:\tAssign the integer num3 to delta_off, default value is 1.\n");
  return TH_FAIL;
}

 /* print the threshold network */
int com_thprint(network_t **network, int argc, char **argv)
{
  int i, c, neg = 0;
  array_t *node_vec;
  
  /* check if threshold network */
  if (is_threshold_network(*network) != TH_SUCCESS)
    return TH_FAIL;

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "dh")) != EOF) {
    switch (c) {
    case 'd':
      neg = 1; break;
    case 'h':
      return thprint_usage();
    default:
      return thprint_usage();
    }
  }
  
  /* process all nodes */
  node_vec = com_get_nodes(*network, argc-util_optind+1, argv+util_optind-1);
  if (array_n(node_vec) < 1)
    return thprint_usage();

  for (i = 0; i < array_n(node_vec); i++)
    thnode_print(sisout, array_fetch(node_t *, node_vec, i), neg);

  /*  don't forget to free space */
  array_free(node_vec);
  return TH_SUCCESS;
}

 /* thprint usage */
static int thprint_usage(void)
{
  fprintf(siserr, "usage: thprint [-d] [n1 n2 ...]\n");
  fprintf(siserr, "    -d \tPrint in negative form\n");
  return TH_FAIL;
}


/* print statistics of thresnold network */
int com_thprint_stats(network_t **network, int argc, char **argv)
{
  array_t *node_vec;
  int c;

  /* check if threshold network */
  if (is_threshold_network(*network) != TH_SUCCESS)
    return TH_FAIL;

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "h")) != EOF) {
    switch (c) {
    case 'h':
      return thprint_stats_usage();
    default:
      return thprint_stats_usage();
    }
  }

  /* process specified nodes */
  if (argc - util_optind >= 1) {
    node_vec = com_get_nodes(*network, argc-util_optind+1, argv+util_optind-1);
    if (array_n(node_vec) < 1)
      return thprint_stats_usage();
    thprint_stats(*network, sisout, node_vec);
    array_free(node_vec);   
  }
  else /*  process the entire network */
    thprint_stats(*network, sisout, NIL(array_t));

  return TH_SUCCESS;
}

/* print statistics */
void thprint_stats(network_t *network, FILE *fp, array_t *node_vec)
{
  int i, levels;
  double area;
  node_t *n;
  lsGen gen;
  array_t *node_array;


  /* particular nodes */
  if (node_vec != NIL(array_t) && array_n(node_vec) > 0) {
    for (i = 0; i < array_n(node_vec); i++) {
      n = array_fetch(node_t *, node_vec, i);
      fprintf(fp, "%s, fanin = %3d, fanout = %3d",
	      node_name(n), node_num_fanin(n), node_num_fanout(n));
      
      fprintf(fp, ", area = %.2f\n", thnode_calc_area(n, NULL));
    }
  }
  /* entire network */
  else {
    area = 0.0;
    foreach_node(network, gen, n)
      if (node_type(n) == INTERNAL) 
	area += thnode_calc_area(n, NULL);
    
    /* compute number of levels  */
    levels = 0;
    node_array = array_alloc(node_t *, 0);
    foreach_primary_output(network, gen, n) {
      array_insert_last(node_t *, node_array, n);
    }
    levels = speed_print_level(network, node_array, 0.5, 0, 0) + 1;

    fprintf(fp, "# of pi: \t %d\n", network_num_pi(network));
    fprintf(fp, "# of po: \t %d\n", network_num_po(network));
    fprintf(fp, "# of levels: \t %d\n", levels);
    fprintf(fp, "# of nodes: \t %d\n", network_num_internal(network));
    fprintf(fp, "area: \t %.2f\n", area);
    fprintf(fp, "network name: %s\n", network_name(network));
  }
}
			  

/* thprint_stats usage */
static int thprint_stats_usage(void)
{
  fprintf(siserr, "usage: thprint_stats [n1 n2 ...]\n");
  return TH_FAIL;
}


 /* check if any inverter exists inside of the given network */
int com_invcheck(network_t **network, int argc, char **argv)
{
  lsGen lgen;
  node_t *n;
  node_t *fanin;
  input_phase_t phase;
  int c, fanin_index, inv_count = 0;

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "h")) != EOF) {
    switch(c) {
    case 'h':
	fprintf(siserr, "Check how namy inverters exists inside of a given network (not including the primary inputs)\n");
	return 0;
    default: break;
    }
  }





  foreach_node(*network, lgen, n) {
    if ((node_type(n) != PRIMARY_INPUT) && (node_type(n) != PRIMARY_OUTPUT)) {
      node_minimum_base(n);
      //      printf("Processing next node:\n");
      //      node_print(sisout, n);
      //      printf("total %d fanins: \n", node_num_fanin(n));
      foreach_fanin(n, fanin_index, fanin) {
	if (node_type(fanin) != PRIMARY_INPUT) {
	  //  node_print(sisout, fanin);
	  if (node_type(fanin) == PRIMARY_OUTPUT) {
	    printf("This node has a PRIMARY_OUTPUT as input\n");
	    node_print(sisout, n);
	  }
	  else {
	    phase = node_input_phase(n, fanin);
	    if (phase == POS_UNATE)
	      ; 
	    else if (phase == NEG_UNATE) {
	      inv_count++;
	      node_print(sisout, n);
	    }
	    else if (phase == BINATE) {
	      inv_count++;
	      node_print(sisout, n);
	    }
	    else
	      fail("unrecognized phase in a node");
	  }
	}
	else {
	  //	  printf("fanin '%s' is a PI\n", node_name(fanin));
	  ;
	}
      }
    }    
  }

  if (inv_count == 0)
    printf("NO inverters inside network '%s'.\n", network_name(*network));
  else
    printf("total %d inverters inside network '%s'.\n", inv_count, network_name(*network));

  return 0;
}
