/*=========================================================================*
 *	File    :	thnode.c
 *
 *	Author  :	Pallav Gupta
 *	Created :	Mon Sep 15 14:18:57 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thnode.c,v 1.12 2004/06/21 18:49:09 sis Exp $ */

#include "sis.h"
#include "th_int.h"

/* default area model */
static double thnode_area_model ARGS((thnode_t *, int));

static void create_key(char *q, thnode_t *th);

 /* allocate a new thnode  */
thnode_t *thnode_alloc(void)
{
  thnode_t *th;

  th = ALLOC(thnode_t, 1);
  th->nin = 0;
  th->W = NIL(double);
  th->T = 0.0;
  th->area = 0.0;

  /* for atpg */
  th->scratch = 0.0;
  th->fanin_lines = NIL(int);
  th->pcubes = NIL(pset_family);
  th->scovers = NIL(pset_family);
  th->simchanged = 0;

  /* everything that is added in thnode_t should have a line in here   */
  return th;
}

/* duplicate a thnode */
thnode_t *thnode_dup(thnode_t *p)
{
  thnode_t *new;
  int i;

  new = thnode_alloc();
  new->nin = p->nin;
  new->W = ALLOC(double, new->nin);

  for (i = 0; i < new->nin; i++)
    new->W[i] = p->W[i];

  new->T = p->T;
  new->area = p->area;

  return new;
}


 /* free an existing thnode */
void thnode_free(thnode_t *th)
{
  int i;

  if (th == NIL(thnode_t))
    return;

  if (th->W != NIL(double)) FREE(th->W);
  
  if (th->fanin_lines != NIL(int))
    FREE(th->fanin_lines);

  if (th->pcubes != NIL(pset_family)) {
    for (i = 0; i < th->nin; i++)
      sf_free(th->pcubes[i]);
    FREE(th->pcubes);
  }

  if (th->scovers != NIL(pset_family)) {
    sf_free(th->scovers[0]);
    sf_free(th->scovers[1]);
  }
  
  /* if any other data strcutures are embedded free in a similar manner */
  FREE(th);
}

/* check if we have a valid threshold node */
int thnode_check(node_t *n)
{
  thnode_t *th;

  th = (thnode_t *) n->th;

  /* Add all check in here */
  /* Right now we only check if the data structure exists,
     In future, we probably wanna check the fields as well. */
  if (th == NIL(thnode_t))
    fail("node does not have a thnode structure");

  return TH_SUCCESS;
}


/* print a threshold node */
void thnode_print(FILE *fp, node_t *n, int phase)
{
  int i;
  node_t *fanin;
  thnode_t *th;
  node_type_t type;

  type = node_type(n);
  
  if ((type != INTERNAL) || 
      (type == PRIMARY_INPUT) || 
      ((type == PRIMARY_OUTPUT) && (node_type(node_get_fanin(n, 0)) /* n->fanin[0]->type*/ == PRIMARY_INPUT)))
    return;

  thnode_check(n);
  th = (thnode_t *) n->th;
  
  //rui
 /*  FILE *fp1, *fp2; */

/*   if ((fp1 = fopen("t_distribute.txt", "a")) == NULL) */
/*     fail("thnode_print: unable to open file handle for writing"); */
/*   if ((fp2 = fopen("w_distribute.txt", "a")) == NULL) */
/*     fail("thnode_print: unable to open file handle for writing"); */

/*   fprintf(fp1, "%6.1f \n", th->T); */
/*   foreach_fanin(n, i, fanin) */
/*     fprintf(fp2, "%6.1f \n", th->W[i]); */

/*   fclose(fp1); */
/*   fclose(fp2); */
  //rui

  fprintf(fp, "%6s =%6.1f; ", node_name(n), th->T);
  foreach_fanin(n, i, fanin)
    fprintf(fp, "%6s =%6.1f; ", node_name(fanin), th->W[i]);
  
  if (phase)
    node_print_negative(fp, n);
  else
    node_print(fp, n);
  //rui
}


/*  attaches a threshod node data structure to this node */
/*  Note: it is assumed that the logic function of n is a simple AND/OR/INV/BUF  */
/*  gate only. It is a serious error to supply anything else. */
void thnode_assign_weights(node_t *n, thsyn_options_t *opt)
{
  int j, sub = 0;
  double *weights;
  thnode_t *th;
  node_t *fanin;
  input_phase_t phase;

  if ((node_function(n) == NODE_AND) || (node_function(n) == NODE_OR)|| 
      (node_function(n) == NODE_INV)|| (node_function(n) == NODE_BUF) ) {
    /* write AMPL data file  */
    thsyn_write_ampl_data(n, opt);
    
    /* system call to AMPL */
    if (system("ampl thsyn_ampl_commands 2>/dev/null") == -1)
      fail("thnode_assign_weights: system call failed.");
    th = thsyn_create_thnode(n, opt, THSYN_NO_COMBINE); 
  }
  else {
    weights = ALLOC(double, node_num_fanin(n));
    
    /*  calculate weights */
    if (node_type(n) == INTERNAL) {
      foreach_fanin(n, j, fanin) {
	phase = node_input_phase(n, fanin);
	
	if (phase == POS_UNATE)
	  weights[j] = 1;
	else if (phase == NEG_UNATE) {
	  weights[j] = -1;
	  sub++;
	}
	else 
	  fail("binate phase in a single node");
      }
    }
    th = thnode_alloc();
    th->nin = node_num_fanin(n);
    th->W = weights;
    
    /* calculate threshold */
    switch (node_function(n)) {
    case NODE_AND:
      th->T = node_num_fanin(n) - sub;
      break;
    case NODE_OR:
      th->T = 1 - sub;
      break;
    case NODE_INV:
      th->T = 0;
      break;
      /* primary inputs and outputs are fed from internal ndoes via buffers */
    case NODE_BUF:
    case NODE_PI:
    case NODE_PO:
      th->W[0] = 1;
      th->T = 1;
      break;
    case NODE_1:
      th->T = 0;
      break;
    case NODE_0:
      th->T = 1;
      break;
    default:
      fail("encountered illegal node type in assigning threshold weights");
    }
  }
  /*  attach the thnode to the node */
  n->th = (char *) th;
}

/* calculate the area of this thnode */
double thnode_calc_area(node_t *n, double (*f)())
{
  thnode_t *th;

  thnode_check(n);
  th = (thnode_t *) n->th;
  /*  if f is null, we use the default area model */
  /*    to calculate the area; */
  if (f == NULL)
    f = thnode_area_model;

  return f(th, node_num_fanin(n));
}


/*  default area model for a thnode */
static double thnode_area_model(thnode_t *th, int nin)
{
  int i;
  double area = fabs(th->T);

  for (i = 0; i < nin; i++)
    area += fabs(th->W[i]);

  return area;  /* sum of all absolute value of weights + threshold */
}


/* simulate a thnode */
void thnode_sim(node_t *n)
{
  thnode_t *th;
  node_t *fanin;
  int i;
  double sum = 0.0;

  thnode_check(n);
  th = (thnode_t *) n->th;
  
  /* be careful, the reason u can't use a for loop is to stay consistent with the 
     variable ordering in the network, so we should always use functions like foreach_____ to get
     the correct weights */
  foreach_fanin(n, i, fanin) {
    sum += (th->W[i] * TH_GETVALUE(fanin));
  }

  /* determine if output is 1 or 0 */
  if (sum >= th->T)
    TH_SETVALUE(n, 1);
  else
    TH_SETVALUE(n, 0);
}

/* simulate a node using 32 vectors compressed in a single integer (bit by bit) */
void thnode_sim32(node_t *n)
{
  thnode_t *th;
  node_t *fanin;
  int i, j;
  unsigned int val, out = 0;
  double sum[TH_SIM32];

  thnode_check(n);
  th = (thnode_t *) n->th;

  //rui
  static double var = 0.0;
  static double mean = 0.0;
  double random_noise;
  //set var and mean values here
  var = 0.0;
  mean = 0.0;

  /* DON'T SIMULATE A PRIMARY INPUT NODE: DO NOT DELETE THIS LINE!! */
  if (node_type(n) == PRIMARY_INPUT)
    return;

  /* initialize */
  for (i = 0; i < TH_SIM32; i++)
    sum[i] = 0.0;
  
  /* be careful, the reason u can't use a for loop is to stay consistent with the 
     variable ordering in the network, so we should always use functions like foreach_____ to get
     the correct weights */
  foreach_fanin(n, i, fanin) {
    val = TH_GETVALUE(fanin);
    /* isolate the j^th bit in the weight and the j^th bit in the val and multiply */
    for (j = 0; j < TH_SIM32; j++) {
      sum[j] += (double) th->W[i] * ((val >> j) & BIT_ZERO);
      //rui 
      random_noise = mean + var *(rand()/RAND_MAX - 0.5);
      sum[j] += random_noise;
      //rui
    }
  }

  /* determine if output is 1 or 0, note the reverse order */
  for (j = TH_SIM32 - 1; j >= 0; j--) {
    //rui
    if (rint(sum[j]) >= rint(th->T))
      out |= BIT_ZERO; /* set the lowest bit */
    //rui
    //    if (sum[j] >= th->T)
    //      out |= BIT_ZERO; /* set the lowest bit */
    if (j > 0)
      out <<= 1; /* shift left */
  }
  
  /* put the 32 bit result in node */
  TH_SETVALUE(n, out);
}

/* create fanin lines for a node */
void thnode_create_fanin_lines(node_t *n)
{
  int i, num;
  thnode_t *th = (thnode_t *) n->th;

  TH_ASSERT(thnode_check(n) == TH_SUCCESS);
  if (th->fanin_lines != NIL(int))
    return;
  
  num = node_num_fanin(n);
  th->fanin_lines = ALLOC(int, num);
  for (i = 0; i < num; i++)
    th->fanin_lines[i] = 0;
}

/* determines all the propagation cubes for a gate for each 
   input and returns an array of of pointers */
pset_family *thnode_pcubes(node_t *n, st_table *pcubes_table)
{
  int i, j, o1, o2, index;
  char *key;
  node_t *fanin;
  thnode_t *th;
  pset_family *f;
  pset q;

  TH_ASSERT(thnode_check(n) == TH_SUCCESS);
  th = (thnode_t *) n->th;
  if (th->pcubes != NIL(pset_family))
    return th->pcubes;

  key = ALLOC(char, TH_MAXWORD);
  create_key(key, th);
  
  /* does this gate exist already, if not we have to calc prop cubes */
  if (!st_lookup(pcubes_table, key, (char **) &f)) {
    f = ALLOC(pset_family, th->nin);
    foreach_fanin(n, index, fanin) {
      f[index] = sf_new(1, 2 * th->nin);

      /* exahaustively simualte 2^inputs combinations */
      for (i = 0; i < (1 << th->nin); i++) {
	for (j = 0; j < th->nin; j++) {
	  if (j == index) continue;
	  TH_SETVALUE(node_get_fanin(n, j), ((i >> j) & 0x1));
	}
	/* set input to 0,1 and simulate */
	TH_SETVALUE(node_get_fanin(n, index), 0);
	thnode_sim(n);
	o1 = TH_GETVALUE(n);
	TH_SETVALUE(node_get_fanin(n, index), 1);
	thnode_sim(n);
	o2 = TH_GETVALUE(n);
	
	/* this is a propogation cube */
	if (o1 != o2) {
	  q = set_new(2 * th->nin);
	  for (j = 0; j < th->nin; j++) {
	    /* this input will serve to tell us what gets propagated */
	    if (j == index) {
	      if (!o1 && o2) /* o1 = 0 and o2 = 1 */
		set_insert(q, 2 * j); /* this means D gets propagated */
	      else
		set_insert(q, 2 * j + 1); /* this means D_BAR gets propagated */
	      continue;
	    }
	    if (TH_GETVALUE(node_get_fanin(n, j)) == 0)
	      set_insert(q, 2 * j);
	    else
	      set_insert(q, 2 * j + 1);
	  }
	  /* add cube to set */
	  f[index] = sf_addset(f[index], q);
	  set_free(q);
	}
      }
      /* remove duplicates etc */
      f[index] = sf_contain(f[index]);
    }

    /* add set to table */
    st_insert(pcubes_table, key, (char *) f);
  }

  FREE(key);
  return f;
}

/* calculate the on and off cubes for each gate */
pset_family *thnode_scovers(node_t *n, st_table *scovers_table)
{
  pset_family *f, fon, foff;
  pset cube;
  char *key;
  thnode_t *th;
  int i, j;

  TH_ASSERT(thnode_check(n) == TH_SUCCESS);
  th = (thnode_t *) n->th;

  if (th->scovers != NIL(pset_family))
    return th->scovers;

  key = ALLOC(char, TH_MAXWORD);
  create_key(key, th);

  /* if entry exists, return it */
  if (!st_lookup(scovers_table, key, (char **) &f)) {
    f = ALLOC(pset_family, 2);

    /* on-set and off-set */
    fon = sf_new(1, 2 * th->nin);
    foff = sf_new(1, 2 * th->nin);

    /* simulate exhaustively */
    for (i = 0; i < (1 << th->nin); i++) {
      for (j = 0; j < (th->nin); j++)
	TH_SETVALUE(node_get_fanin(n, j), ((i >> j) & 0x1));
      thnode_sim(n);
      
      cube = set_new(2 * th->nin);
      for (j = 0; j < th->nin; j++) {
	if (TH_GETVALUE(node_get_fanin(n, j)) == 0)
	  set_insert(cube, 2 * j);
	else
	  set_insert(cube, 2 * j + 1);
      }

      /* add to on-set or off-set */
      if (TH_GETVALUE(n))
	fon = sf_addset(fon, cube);
      else
	foff = sf_addset(foff, cube);
      
      set_free(cube);
    }
    
    /* merge the cubes to form large cubes */
    for (i = 0; i < th->nin; i++) {
      fon = d1merge(fon, i);
      foff = d1merge(foff, i);
    }
    f[0] = sf_contain(fon);
    f[1] = sf_contain(foff);

    st_insert(scovers_table, key, (char *) f);
  }

  FREE(key);
  return f;
}

/* form the hash key, just concanatenate all the weigts/threshold into string */ 
static void create_key(char *q, thnode_t *th)
{
  int i;

  /* use only 2 precision digits, otherwise strcmp goes beserk during hashing */
  for (i = 0; i < th->nin; i++)
    q += sprintf(q, "%.2lf", th->W[i]); 
  sprintf(q, "%.2lf", th->T);
}
