/*=========================================================================*
 *	File    :	thvcg.c
 *
 *	Author  :	Rui Zhang
 *	Created :	Thu Sep 25 23:21:02 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thvcg.c,v 1.7 2004/01/16 04:07:39 pgupta Exp $ */

#include "sis.h"
#include "th_int.h"

static int  thvcg_usage ARGS(());
static void thvcg_dump ARGS((FILE *, network_t *, array_t *, thvcg_options_t *));
static void thvcg_graph_print ARGS((FILE *, network_t *, thvcg_options_t *));
static void thvcg_node_print ARGS((FILE *, node_t *, int, thvcg_options_t *));
static void thvcg_edge_print ARGS((FILE *, array_t *node_vec, node_t *, int, thvcg_options_t *));
static array_t *thvcg_find_fanin ARGS((node_t *, array_t *, int));
static void thvcg_single_edge ARGS((FILE *, int, int, int, thvcg_options_t *));

int com_thvcg(network_t **network, int argc, char **argv)
{
  FILE *opf;
  array_t *node_vec;
  thvcg_options_t *opt;

  /* options parse and checks */
  if ((opt = thvcg_options_init(argc, argv)) == NIL(thvcg_options_t)
      || opt->help) {
    thvcg_options_free(opt);
    return thvcg_usage();
  }  
  
  /* check if threshold network */
  if (is_threshold_network(*network) != TH_SUCCESS) {
    thvcg_options_free(opt);
    return TH_FAIL;
  }
  
  if ((opf = com_open_file(opt->filename, "w", 0, 0)) == NIL(FILE)) {
    fprintf(siserr, "Unable to open file '%s' to write\n", opt->filename);	
    thvcg_options_free(opt);
    return TH_FAIL;
  }

  //  opf = fopen(opt->filename, "w");
  //if (opf == NULL) {
  //  fprintf(siserr, "Unable to open file '%s' to write\n", opt->filename);	
  //  FREE(opt->filename);
  //}  

  node_vec = network_dfs(*network);
  //  node_vec = com_get_nodes(*network, argc-util_optind+1, argv+util_optind-1);
  if (array_n(node_vec) < 1)
    return thvcg_usage();
  thvcg_dump(opf, *network, node_vec, opt);
  /*  don't forget to free space */
  array_free(node_vec);
  fclose(opf);

  /* free options */
  thvcg_options_free(opt);

  return TH_SUCCESS;
}


/* dump vcg formats into the file *fp */
static void thvcg_dump(FILE *fp, network_t *network, array_t *node_vec, thvcg_options_t *opt)
{ 
  int i; 
  node_t *n;

  /* printing graph description */
  thvcg_graph_print(fp,network,opt);

  /* printing node and edge descriptions */
  for (i = 0; i < array_n(node_vec); i++) {
    n = array_fetch(node_t *, node_vec,i);
    if (node_type(n) != PRIMARY_OUTPUT) {
      /* printing node description */
      thvcg_node_print(fp, n, i, opt);
      /* printing edge description */
      thvcg_edge_print(fp, node_vec, n, i, opt);
    }
  }
  /* print end of graph */
  fprintf(fp,"       }\n");
}

static void thvcg_graph_print(FILE *fp, network_t *network, thvcg_options_t *opt)
{
  /* print graph attributes */  
  fprintf(fp,"graph: { title: \"%s\" \n", network->net_name);
  fprintf(fp,"         label: \"%s\" \n", network->net_name);
  fprintf(fp,"         layoutalgorithm: dfs \n");
  fprintf(fp,"         orientation: left_to_right \n");
  fprintf(fp,"         display_edge_labels: yes \n");
  /* print node attributes */
  fprintf(fp,"         node.shrink: %d \n", opt->shrink);
  fprintf(fp,"         node.textmode: center \n");
  fprintf(fp,"         node.shape: ellipse \n");
  fprintf(fp,"         node.color: lightgreen node.textcolor: black \n");
  /* print edge attributes */
  fprintf(fp,"         edge.color: red  edge.priority: 1 \n");
}

static void thvcg_node_print(FILE *fp, node_t *n, int index, thvcg_options_t *opt)
{ 
  int i;
  node_t *fanin;
  node_type_t type;
  thnode_t *th;

  type = node_type(n);
  fprintf(fp,"node: { title: \"%d\" ", index);
  if (type == INTERNAL) { 
    if ((type == PRIMARY_INPUT) || (type == PRIMARY_OUTPUT))
      fprintf(fp,"label: \"%s\"", node_name(n));
    else {	  
      fprintf(fp,"label: \"");
      th = (thnode_t *) n->th;
      fprintf(fp, "%s", node_name(n));
      if (opt->mode == 1) // printing node according to mode 1 format 
	{
	  fprintf(fp, "\\n");
	  fprintf(fp,"%d\"",(int)th->T);
	}
      else // printing node according to mode 2 format
	{
	  fprintf(fp, "; ");
	  foreach_fanin(n, i, fanin)
	    fprintf(fp, "%s,", node_name(fanin));
	  fprintf(fp, "\\n");
	  fprintf(fp,"%d;",(int)th->T);
	  foreach_fanin(n, i, fanin)
	    fprintf(fp, "%d,", (int)th->W[i]);
	  fprintf(fp,"\"");
	}
    }
  }
  else 
    fprintf(fp,"label: \"%s\"", node_name(n));
  
  fprintf(fp,"}\n");
}

static void thvcg_edge_print(FILE *fp, array_t *node_vec, node_t *n, int index, thvcg_options_t *opt)
{
 int i, j, w, k;
 node_t *fanin, *source_n, *n_copy, *nin;
 node_type_t type;
 array_t *fi_vec;
 thnode_t *th;

 n_copy = node_dup(n);
 type = node_type(n);	      
 th = (thnode_t *) n->th;

 if (type != PRIMARY_INPUT) {
   if (type == PRIMARY_OUTPUT) {
     fi_vec = thvcg_find_fanin(n, node_vec, index);
     if (array_n(fi_vec) > 0)
       for(j = 0; j < array_n(fi_vec); j++)
	 thvcg_single_edge(fp, array_fetch(int, fi_vec, j), index, 0, opt);
   }
   else {
     foreach_fanin(n, i, fanin) {
       fi_vec = thvcg_find_fanin(fanin, node_vec, index);
       if (array_n(fi_vec) > 0)
	 for(j = 0; j < array_n(fi_vec); j++) {
	   // find out the weight
	   // fetch out the source node
	   source_n = array_fetch(node_t *, node_vec, array_fetch(int, fi_vec, j));
	   foreach_fanin(n_copy, k, nin) {	
	     // find out the index of the weight
	     if(strcmp(node_name(source_n), node_name(nin)) == 0)
	       // get the corresponding weight
	       w = (int)th->W[k]; 
	   }
	   thvcg_single_edge(fp, array_fetch(int, fi_vec, j), index, w, opt);
	 }
     }
   }
 }
 FREE(n_copy);
}

/* find all the indexes of the nodes that are the fanout of the given node name */
static array_t *thvcg_find_fanin(node_t *des_n, array_t *node_vec, int index)
{
  int i;
  node_t *tmp_n;
  array_t *fi_nodes;  
  
  /* allocate the array to save the fanin nodes' index */
  fi_nodes = array_alloc(int,0);
  for (i = 0; i < array_n(node_vec); i++)
    if ( i != index)  {  
      tmp_n = array_fetch(node_t *,node_vec,i);
      /* find a fanout node */
      if(strcmp(node_name(des_n), node_name(tmp_n)) == 0) {
	// save this index i into a array_t *fi_nodes
	array_insert_last(int, fi_nodes, i);
      }
    }
  return fi_nodes;
}

/* given the source and destination of each node, print out the 
 * corresponding single edge
 */
static void thvcg_single_edge(FILE *fp, int source, int desti, int weight, thvcg_options_t *opt)
{
  if (opt->mode == 1) // printing node according to mode 1 format 
    fprintf(fp,"edge: { sourcename:\"%d\" targetname:\"%d\" label: \"%d\"}\n", source, desti, weight);
  else
    fprintf(fp,"edge: { sourcename:\"%d\" targetname:\"%d\"}\n", source, desti);
}

/* initialize thvcg options data structrure */
thvcg_options_t *thvcg_options_init(int argc, char **argv)
{
  int c;
  thvcg_options_t *opt;
  
   opt = ALLOC(thvcg_options_t, 1);
   /* default values */
   opt->mode = 1;
   opt->shrink = 1;
   opt->filename = util_strsav("thvcg_output.vcg");
   opt->help = FALSE;

  /* parse options */
  util_getopt_reset();
 while ((c = util_getopt(argc, argv, "o:m:s:h")) != EOF) {
    switch (c) {
    case 'o':
      opt->filename = util_strsav(util_optarg);
      break;
    case 'm':
      opt->mode = atoi(util_optarg);
      if (!((opt->mode == 1) || (opt->mode == 2)))
	opt->help = TRUE;
      break;
    case 's':
      opt->shrink = atoi(util_optarg);
      break;
    case 'h':
      opt->help = TRUE; break;
    default:
      opt->help = TRUE; break;
    }
 }
 return opt;
}


/* free options data structure */
void thvcg_options_free(thvcg_options_t *opt)
{
  if (opt == NIL(thvcg_options_t))
    return;
  if (opt->filename != NIL(char))
    FREE(opt->filename);
  FREE(opt);
}


/* thvcg usage */
static int thvcg_usage(void)
{
  // do the usage information here
  fprintf(siserr, "usage: thvcg [-o file] [-m int1] [-s int2] [-h]\n");
  fprintf(siserr, "    -o <file> \tWrite the vcg file to <file>, default write to thvcg_output.vcg.\n");
  fprintf(siserr, "    -m <int1> \tAssign the printing mode to int1. Print weight on the edge when int1=1.\n");
  fprintf(siserr, "              \tPrint weight-threshold vector inside of the node when int1=2\n");
  fprintf(siserr, "              \t(default is 1).\n");
  fprintf(siserr, "    -s <int2> \tShrink the node size into (1/int2)*100 percent (default is 1).\n");
  fprintf(siserr, "    -h        \tPrint out help information\n");
  return TH_FAIL;
}
